from .probelib.NetDataProbe import NetDataProbe as DefaultMonitoring
from .probelib.AppMetricsProbe import AppMetricsProbe as FileUserDefinedMetrics
from .probelib.ServerAppMetricsProbe import ServerAppMetricsProbe as UserDefinedMetrics
from .probelib.WebAppMetricsProbe import WebAppMetricsProbe as WebUserDefinedMetrics
from .probelib.DockerProbe import DockerProbe as ContainerMetrics
from .utils import RainbowUtils, NetDataController
from .Probe import Probe
from .Metric import Metric, DiffMetric, CounterMetric, SimpleMetric, TimerMetric




import docker
from RainbowMonitoringSDK.probes.Probe import Probe
from RainbowMonitoringSDK.probes.Metric import SimpleMetric

class DockerProbe(Probe):

    def __init__(self,
                 periodicity, debug=False, _logging=False,
                 docker_client=docker.from_env(),
                 **kwargs):
        Probe.__init__(self, "DockerProbe", periodicity,  debug, _logging)
        self.docker_client = docker_client
        self.config = kwargs

    def get_desc(self):
        return "The monitoring probe of docker containers"

    def collect(self):
        raw_metrics = self.retrieve_metrics()
        for raw_metric in raw_metrics:
            metric = self.get_metric(raw_metric.get_name())
            if not metric:
                metric = raw_metric
                self.add_metric(metric)
            metric.set_val(raw_metric.get_val())
            metric.set_group('DOCKER')  # TODO Change to be more group specific

    def retrieve_metrics(self) -> [SimpleMetric]:
        containers = self.docker_client.containers.list()
        metrics = []
        pod_stats = {}
        for container in containers:
            curr_metrics = []
            attrs = container.attrs
            attributes = container.attrs.get('Config', {})
            stats = container.stats(stream=False)

            is_pod = attributes.get("Labels", {}).get("io.kubernetes.container.name") == "POD"
            
            if is_pod: continue
            container_id = attrs.get("Id")[:12]
            pod_uid = attributes.get("Labels", {}).get("io.kubernetes.pod.uid")
            pod_name = attributes.get("Labels", {}).get("io.kubernetes.pod.name")
            container_name = attributes.get("Labels", {}).get("io.kubernetes.container.name")
            pod_namespace = attributes.get("Labels", {}).get("io.kubernetes.pod.namespace")
            prefix = f"CONTAINER_{container_id}"
            if pod_uid:
                prefix = f"CONTAINER_{pod_namespace}|{pod_name}|{pod_uid}|{container_name}|{container_id}|"
            curr_metrics.append(self.get_cpu_ms(stats, prefix))
            curr_metrics.append(self.get_cpu_ptc(stats, prefix))
            curr_metrics.append(self.get_memory_usage(stats, prefix))
            curr_metrics.append(self.get_memory_usage_ptc(stats, prefix))

            all_net_metrics = {}

            for eth, net_stats in stats.get('networks', {}).items():
                for i in net_stats:
                    all_net_metrics[i] = all_net_metrics.get(i, 0) + net_stats[i]
                curr_metrics.extend(self.get_network_metrics(net_stats, eth, prefix, stats['read']))

            curr_metrics.extend(self.get_network_metrics(all_net_metrics, "overall_network", prefix, stats['read']))

            if pod_uid:
                pod_prefix = f"POD_{pod_namespace}|{pod_name}|{pod_uid}|"
                cpu_ms = self.get_cpu_ms(stats, pod_prefix)
                cpu_ptc = self.get_cpu_ptc(stats, pod_prefix)
                memory_usage = self.get_memory_usage(stats, pod_prefix)
                memory_usage_ptc = self.get_memory_usage_ptc(stats, pod_prefix)
                self.update_pod_stats(cpu_ms, pod_stats)
                self.update_pod_stats(cpu_ptc, pod_stats)
                self.update_pod_stats(memory_usage, pod_stats)
                self.update_pod_stats(memory_usage_ptc, pod_stats)
                network_usage = self.get_network_metrics(all_net_metrics, "overall_network", pod_prefix, stats['read'])
                for net in network_usage:
                    self.update_pod_stats(net, pod_stats)
            metrics.extend(curr_metrics)
        for _, metric in pod_stats.items():
            metrics.append(metric)
        return metrics

    def update_pod_stats(self, curr_metric, pod_stats):
        pod_curr_metric = pod_stats.get(curr_metric.get_name(),
                                   SimpleMetric(curr_metric.get_name(), curr_metric.get_units(), curr_metric.get_desc(),
                                                higherIsBetter=curr_metric.get_higherisbetter()))
        value = pod_curr_metric.get_val()
        pod_curr_metric.set_val(value + curr_metric.get_val() if value else curr_metric.get_val())
        pod_stats[curr_metric.get_name()] = pod_curr_metric

    def get_network_metrics(self, stats, net, prefix, timestamp):
        metrics = []
        for i in stats:
            usage = stats[i]
            simple_metric = SimpleMetric(
                f"{prefix}_{net}_{i}", i.split("_")[1], f"Network metric ({i}) of the container for {net} interface", minVal=0,
                                         higherIsBetter=False)

            simple_metric.set_val(usage)
            simple_metric.set_timestamp(timestamp)
            metrics.append(simple_metric)

        return metrics

    def get_memory_usage(self, stats, prefix=""):
        try:
            usage = stats['memory_stats']['usage']
        except Exception:
            usage = 0
        simple_metric = SimpleMetric(f"{prefix}_memory", "bytes", "Occupied memory of the container", minVal=0, higherIsBetter=False)

        simple_metric.set_val(usage)
        simple_metric.set_timestamp(stats['read'])
        return simple_metric

    def get_memory_usage_ptc(self, stats, prefix=""):
        try:
            usage = stats['memory_stats']['usage']
            limit = stats['memory_stats']['limit']
            percentage = usage/limit
            percent = round(percentage, 5)
        except Exception:
            percent = 0
        simple_metric = SimpleMetric(f"{prefix}_memory_ptc", "%", "Percentage memory of the container", minVal=0,
                                     higherIsBetter=False)
        simple_metric.set_val(percent)
        simple_metric.set_timestamp(stats['read'])
        return simple_metric

    def get_cpu_ms(self, stats, prefix=""):
        try:
            total_cpu_usage = stats['cpu_stats']['cpu_usage']['total_usage']
        except Exception:
            total_cpu_usage = 0
        simple_metric = SimpleMetric(f"{prefix}_cpu", "ms", "cpu runtime of the container", minVal=0,
                                     higherIsBetter=False)
        simple_metric.set_val(total_cpu_usage)
        simple_metric.set_timestamp(stats['read'])
        return simple_metric

    def get_cpu_ptc(self, stats, prefix=""):
        try:
            UsageDelta = stats['cpu_stats']['cpu_usage']['total_usage'] - stats['precpu_stats']['cpu_usage']['total_usage']
            SystemDelta = stats['cpu_stats']['system_cpu_usage'] - stats['precpu_stats']['system_cpu_usage']
            len_cpu = len(stats['cpu_stats']['cpu_usage']['percpu_usage'])
            percentage = (UsageDelta / SystemDelta) * len_cpu * 100
            percent = round(percentage, 2)
        except Exception:
            percent = 0
        simple_metric = SimpleMetric(f"{prefix}_cpu_ptc", "%", "cpu percent utilization of the container", minVal=0, higherIsBetter=False)
        simple_metric.set_val(percent)
        simple_metric.set_timestamp(stats['read'])
        return simple_metric

    def get_metrics(self):
        self.collect()
        return Probe.get_metrics(self)



import re
import subprocess
import requests
import yaml
from yaml.loader import SafeLoader
from RainbowMonitoringSDK.probes.Metric import SimpleMetric
from RainbowMonitoringSDK.probes.Probe import Probe
from RainbowMonitoringSDK.probes.utils import NetDataController


class NetDataProbe(Probe):
    # overall cpu, memomory, disk, disk i/o
    __REGEX_METRIC_NAME_FOR_DEMO = {
        '_SYSTEM_CPU_IDLE': 'cpu_idle',
        '_SYSTEM_CPU_VISIBLETOTAL': 'cpu_ptc',
        '_APPS_SOCKETS_VISIBLETOTAL': '',
        '_SYSTEM_RAM_FREE': 'memory_free',
        '_SYSTEM_RAM_USED': 'memory_used',
        '_SYSTEM_RAM_CACHED': 'memory_cached',
        '_SYSTEM_RAM_BUFFERS': 'memory_buffers',
        '_SYSTEM_RAM_VISIBLETOTAL': 'memory_total',

        '_IPV6_PACKETS_RECEIVED': 'ipv6_packets_received',  # packets/s
        '_IPV6_PACKETS_SENT': 'ipv6_packets_sent',  # packets/s
        '_IPV6_PACKETS_FORWARDED': 'ipv6_packets_forwarded',  # packets/s
        '_IPV6_PACKETS_DELIVERS': 'ipv6_packets_delivers',  # packets/s
        '_IPV6_PACKETS_VISIBLETOTAL': 'ipv6_packets_total',  # packets/s

        '_IPV4_PACKETS_RECEIVED': 'ipv4_packets_received',  # packets/s
        '_IPV4_PACKETS_SENT': 'ipv4_packets_sent',  # packets/s
        '_IPV4_PACKETS_FORWARDED': 'ipv4_packets_forwarded',  # packets/s
        '_IPV4_PACKETS_DELIVERS': 'ipv4_packets_delivers',  # packets/s
        '_IPV4_PACKETS_VISIBLETOTAL': 'ipv4_packets_total',  # packets/s

        '_SYSTEM_IPV6_RECEIVED': 'ipv6_kb_received', # kilobits/s
        '_SYSTEM_IPV6_SENT': 'ipv6_kb_sent',  # kilobits/s
        '_SYSTEM_IPV6_VISIBLETOTAL': 'ipv6_kb_total',  # kilobits/s
        '_SYSTEM_IP_RECEIVED': 'ipv4_kb_received' , # kilobits/s
        '_SYSTEM_IP_SENT': 'ipv4_kb_sent',  # kilobits/s
        '_SYSTEM_IP_VISIBLETOTAL': 'ipv4_kb_total',  # kilobits/s
        '_DISK_SPACE___AVAIL': 'disk_available',
        '_DISK_SPACE___USED': 'disk_used',
        '_DISK_SPACE___RESERVED_FOR_ROOT': 'disk_reserved_for_root',
        '_DISK_SPACE___VISIBLETOTAL': 'disk_total'
        }
    __NETDATA_URL_TEMPLATE = "http://%s:%s/api/v1/allmetrics"

    __NETDATA_STARTED = 0

    def __init__(self,
                 periodicity,
                 prefix="",
                 ip_address='0.0.0.0',
                 port='19999',
                 debug=False,
                 logging=False,
                 *args,
                 **kwargs):
        Probe.__init__(self, "NetData", periodicity, debug, logging)
        self.controller = NetDataController()
        self.ip_address = ip_address
        self.port = port
        self.url = self.__NETDATA_URL_TEMPLATE % (self.ip_address, self.port)
        self.prefix = prefix

    def get_desc(self):
        return "The monitoring probe of netdata"

    def collect(self):
        raw_metrics = self.retrieve_metrics()
        for raw_metric in raw_metrics:
            high_level_name = self.__is_in_demo_metrics(raw_metric['name'])
            if not high_level_name: continue
            metric = self.get_metric(high_level_name)
            if not metric:
                metric = SimpleMetric(high_level_name, raw_metric['unit'], raw_metric['name'])
                self.add_metric(metric)
            metric.set_val(raw_metric['value'])
            metric.set_group('NETDATA')  # TODO Change to be more group specific

    def __is_in_demo_metrics(self, metric_name):
        for regex in self.__REGEX_METRIC_NAME_FOR_DEMO:
            if re.search(regex, metric_name):
                return self.__REGEX_METRIC_NAME_FOR_DEMO.get(regex)
        return False

    def retrieve_metrics(self):
        res = []
        try:
            metrics = requests.get(self.url).text
            self.__NETDATA_STARTED = 1
        except Exception as e:
            print("NetData is not working: ", e)
            return []
        for metric_line in metrics.splitlines():
            name, value, unit = self.__extract_information_from_line(metric_line)
            try:
                float(value)
            except Exception:
                continue
            res.append(dict(name=name, value=value, unit=unit))
        return res

    def __extract_information_from_line(self, metric_line):
        try:
            value_and_unit = metric_line.split("#")
            if len(value_and_unit) == 1:
                value, unit = value_and_unit[0], None
            else:
                value, unit = value_and_unit[0], value_and_unit[1]

            unit = unit.strip().replace(" ", "_")
            value = value.strip().split("=")
            name, value = value[0], value[1].replace('"', "")
            name = name.replace('NETDATA', self.prefix)
            return name, value, unit
        except Exception:
            return None, None, None

    def activate(self):
        # Open the file and load the file
        with open('/code/configs.yaml') as f:
            data = yaml.load(f, Loader=SafeLoader)
            self.controller.update_configs(data["sensing-units"])

        self.a_child_process = subprocess.Popen(["bash", "/usr/sbin/run.sh", "-c", "/code/RainbowMonitoringSDK/probes/net-data.cfg"], shell=False,
                                                stdout=open("netdata-stdout.txt", "w"),
                                                stderr=open("netdata-stderr.txt", "w"))
        Probe.activate(self)

    def cleanup(self):
        self.a_child_process.terminate()


def main():
    p = NetDataProbe("netdata", 5, ip_address="0.0.0.0", port=19999)
    p.set_debugmode(True)
    p.set_logging()

    p.activate()


if __name__ == "__main__":
    main()



import threading
from copy import deepcopy

from RainbowMonitoringSDK.probes.Metric import SimpleMetric
from RainbowMonitoringSDK.probes.Probe import Probe
from flask import request, Flask


class ServerAppMetricsProbe(Probe):

    latest_metrics = []
    app = Flask(__name__)
    def __init__(self,
                 periodicity, debug=False, _logging=False,
                 **kwargs):
        Probe.__init__(self, "ServerAppMetricsProbe", periodicity,  debug, _logging )
        self.config = kwargs
        threading.Thread(target=ServerAppMetricsProbe.start_stand_by_server, args=[self]).start()

    def get_desc(self):
        return "The monitoring probe of user defined metrics"

    def collect(self):
        json_metrics = self.retrieve_metrics()
        for _metric in json_metrics:
            metric = self.__metric_from_dictionary(_metric)
            self.add_metric(metric)

    @staticmethod
    def __metric_from_dictionary(collected):
        metric = SimpleMetric(collected.get('name'), collected.get('units'), collected.get('desc'))
        metric.set_val(collected.get('val'))
        metric.set_group(collected.get('group'))
        return metric

    def retrieve_metrics(self):
        res = deepcopy(self.latest_metrics)
        self.latest_metrics = []
        return res

    def get_metrics(self):
        self.collect()
        return Probe.get_metrics(self)

    def start_stand_by_server(self):
        path = self.config.get("path", "/")
        port = self.config.get("port", "1090")
        @self.app.route(path, methods=["POST"])
        def main_route():
            data = request.json
            if type(data) != dict: return {"success": False}
            for param in ['name', 'units', 'desc', 'val', 'group']:
                if data.get(param) == None: return {"success": False}

            self.latest_metrics.append(data)
            return {"success": True}
        self.app.run(host="0.0.0.0", port=port)